@@ -1,3 +1,4 @@
+from scipy.stats import binned_statistic
 import hnswlib
 import typing
 from numba import njit, prange
@@ -264,7 +265,9 @@ class SAMAP(object):
         ncpus=os.cpu_count(),
         hom_edge_thr=0,
         hom_edge_mode = "pearson",
-        scale_edges_by_corr = False
+        scale_edges_by_corr = False,
+        neigh_from_key1 = False,
+        neigh_from_key2 = False
         
     ):
         """Runs the SAMap algorithm.
@@ -310,7 +313,15 @@ class SAMAP(object):
         scale_edges_by_corr: bool, optional, default False
             If True, scale cell-cell cross-species edges by their expression similarities
             (correlations).
-
+            
+        neigh_from_key1 : bool, optional, default False
+            If True, species 1 neighborhoods are calculated directly from the chosen clustering (`self.key1`).
+            Cells within the same cluster belong to the same neighborhood.
+            
+        neigh_from_key2 : bool, optional, default False
+            If True, species 2 neighborhoods are calculated directly from the chosen clustering (`self.key2`).
+            Cells within the same cluster belong to the same neighborhood.
+            
         Returns
         -------
         samap - Species-merged SAM object
@@ -336,7 +347,9 @@ class SAMAP(object):
             ncpus=ncpus,
             THR=hom_edge_thr,
             corr_mode=hom_edge_mode,
-            scale_edges_by_corr = scale_edges_by_corr
+            scale_edges_by_corr = scale_edges_by_corr,
+            neigh_from_key1=neigh_from_key1,
+            neigh_from_key2=neigh_from_key2
         )
         samap = smap.final_sam
         self.samap = samap
@@ -571,8 +584,8 @@ class SAMAP(object):
         else:
             return self.SamapGui.SamPlot
         
-    def refine_homology_graph(self, THR=0, NCLUSTERS=1, ncpus = os.cpu_count(), corr_mode='pearson'):
-        return self.smap.refine_homology_graph(NCLUSTERS=NCLUSTERS, ncpus=ncpus, THR=THR, corr_mode=corr_mode)
+    def refine_homology_graph(self, THR=0, NCLUSTERS=1, ncpus = os.cpu_count(), corr_mode='pearson', ct_labels = None):
+        return self.smap.refine_homology_graph(NCLUSTERS=NCLUSTERS, ncpus=ncpus, THR=THR, corr_mode=corr_mode, ct_labels=ct_labels)
         
 class _Samap_Iter(object):
     def __init__(
@@ -598,7 +611,7 @@ class _Samap_Iter(object):
         ]
         self.iter = 0
 
-    def refine_homology_graph(self, NCLUSTERS=1, ncpus=os.cpu_count(), THR=0, corr_mode='pearson'):
+    def refine_homology_graph(self, NCLUSTERS=1, ncpus=os.cpu_count(), THR=0, corr_mode='pearson', ct_labels=None):
         if corr_mode=='mutual_info':
             try:
                 from fast_histogram import histogram2d
@@ -627,11 +640,13 @@ class _Samap_Iter(object):
             T2=0,
             NCLUSTERS=NCLUSTERS,
             ncpus=ncpus,
-            corr_mode=corr_mode
+            corr_mode=corr_mode,
+            ct_labels=ct_labels
         )
         return gnnmu
 
-    def run(self, NUMITERS=3, NOPs1=0, NOPs2=0, NH1=2, NH2=2, K=20, corr_mode='pearson', NCLUSTERS=1, scale_edges_by_corr=False, THR=0, ncpus=os.cpu_count()):
+    def run(self, NUMITERS=3, NOPs1=0, NOPs2=0, NH1=2, NH2=2, K=20, corr_mode='pearson', NCLUSTERS=1,
+                  scale_edges_by_corr=False, THR=0, neigh_from_key1=False, neigh_from_key2=False, ncpus=os.cpu_count()):
         sam1 = self.sam1
         sam2 = self.sam2
         gnnm = self.gnnm
@@ -666,7 +681,9 @@ class _Samap_Iter(object):
                 coarsen=True,
                 key1=self.key1,
                 key2=self.key2,
-                scale_edges_by_corr=scale_edges_by_corr
+                scale_edges_by_corr=scale_edges_by_corr,
+                neigh_from_key1=neigh_from_key1,
+                neigh_from_key2=neigh_from_key2
             )
 
             self.samap = sam4
@@ -716,6 +733,16 @@ def _replace(X,Y,xi,yi):
         data[i] = ((x-x.mean())*(y-y.mean()) / x.std() / y.std()).sum() / x.size
     return data
     
+    
+def _generate_coclustering_matrix(cl):
+    cl = ut.convert_annotations(np.array(list(cl)))
+    clu,cluc=np.unique(cl,return_counts=True)    
+    v = np.zeros((cl.size,clu.size))
+    v[np.arange(v.shape[0]),cl]=1
+    v = sp.sparse.csr_matrix(v)
+    return v
+
+    
 def _mapper(
     sams,
     gnnm,
@@ -731,6 +758,8 @@ def _mapper(
     key1="leiden_clusters",
     key2="leiden_clusters",
     scale_edges_by_corr=False,
+    neigh_from_key1=False,
+    neigh_from_key2=False,
     **kwargs
 ):
     n = len(sams)
@@ -772,14 +801,21 @@ def _mapper(
                 print("Out-neighbor smart expansion 1")
                 nnm = sam.adata.obsp["connectivities"].copy()
                 nnm1_out = nnm
-                nnm1_in = _smart_expand(nnm, K1, NH=NH1)
-                nnm1_in.data[:] = 1
+                if not neigh_from_key1:
+                    nnm1_in = _smart_expand(nnm, K1, NH=NH1)
+                    nnm1_in.data[:] = 1
+                else:
+                    nnm1_in = _generate_coclustering_matrix(CL1)
 
                 print("Out-neighbor smart expansion 2")
                 nnm = sam2.adata.obsp["connectivities"].copy()
                 nnm2_out = nnm
-                nnm2_in = _smart_expand(nnm, K2, NH=NH2)
-                nnm2_in.data[:] = 1
+                if not neigh_from_key2:
+                    nnm2_in = _smart_expand(nnm, K2, NH=NH2)
+                    nnm2_in.data[:] = 1
+                else:
+                    nnm2_in = _generate_coclustering_matrix(CL2)                    
+
 
                 mdata["nnm1_out"] = nnm1_out
                 mdata["nnm1_in"] = nnm1_in
@@ -809,19 +845,29 @@ def _mapper(
                 for bl in range(numiter):
                     print(str(bl) + "/" + str(numiter), D.shape, R)
                     if not R:
-                        C = nnm2_in[bl * chunksize : (bl + 1) * chunksize].dot(B.T)
+                        if neigh_from_key2:
+                            C = nnm2_in[bl * chunksize : (bl + 1) * chunksize].dot(nnm2_in.T.dot(B.T))
+                        else:
+                            C = nnm2_in[bl * chunksize : (bl + 1) * chunksize].dot(B.T)
                         C.data[C.data < 0.1] = 0
                         C.eliminate_zeros()
-
-                        C2 = B2[bl * chunksize : (bl + 1) * chunksize].dot(nnm1_in.T)
+                        if neigh_from_key1:
+                            C2 = nnm1_in.dot(nnm1_in.T.dot(B2[bl * chunksize : (bl + 1) * chunksize].T)).T
+                        else:
+                            C2 = nnm1_in.dot(B2[bl * chunksize : (bl + 1) * chunksize].T).T
                         C2.data[C2.data < 0.1] = 0
                         C2.eliminate_zeros()
                     else:
-                        C = B[bl * chunksize : (bl + 1) * chunksize].dot(nnm2_in.T)
+                        if neigh_from_key2:
+                            C = nnm2_in.dot(nnm2_in.T.dot(B[bl * chunksize : (bl + 1) * chunksize].T)).T
+                        else:
+                            C = nnm2_in.dot(B[bl * chunksize : (bl + 1) * chunksize].T).T
                         C.data[C.data < 0.1] = 0
                         C.eliminate_zeros()
-
-                        C2 = nnm1_in[bl * chunksize : (bl + 1) * chunksize].dot(B2.T)
+                        if neigh_from_key1:
+                            C2 = nnm1_in[bl * chunksize : (bl + 1) * chunksize].dot(nnm1_in.T.dot(B2.T))
+                        else:
+                            C2 = nnm1_in[bl * chunksize : (bl + 1) * chunksize].dot(B2.T)
                         C2.data[C2.data < 0.1] = 0
                         C2.eliminate_zeros()
 
@@ -836,7 +882,7 @@ def _mapper(
 
                 if not R:
                     D = D.T
-                    D = D.tocsr()
+                D = D.tocsr()
 
                 mdata["xsim"] = D
                 if scale_edges_by_corr:
@@ -947,6 +993,7 @@ def _refine_corr(
     T2=0,
     NCLUSTERS=1,
     ncpus=os.cpu_count(),
+    ct_labels=None
 ):
     # import networkx as nx
     gn = np.append(gn1, gn2)
@@ -983,6 +1030,7 @@ def _refine_corr(
             T1=T1,
             T2=T2,
             ncpus=ncpus,
+            ct_labels=ct_labels
         )
         GNNMSUBS.append(gnnm2_sub)
         CORRSUBS.append(CORR_sub)
@@ -1384,7 +1432,7 @@ def _get_pairs(sam1, sam2, gnnm, gn1, gn2, NOPs1=2, NOPs2=5):
     # gnnm = filter_gnnm(gnnm)
     su = gnnm.max(1).A
     su[su == 0] = 1
-    gnnm = gnnm.multiply(1 / su).tocsr()
+    #gnnm = gnnm.multiply(1 / su).tocsr()
     W1 = sam1.adata.var["weights"][gn1].values
     W2 = sam2.adata.var["weights"][gn2].values
     W = np.append(W1, W2)
@@ -1425,7 +1473,7 @@ def _avg_as(s):
     )  / s.adata.uns['mdata']['knn_1v2'][0].data.size )
 
 
-def _parallel_init(ipl1x, isc1x, ipairs, ign1O, ign2O, iT2, iCORR, icorr_mode):
+def _parallel_init(ipl1x, isc1x, ipairs, ign1O, ign2O, iT2, iCORR, icorr_mode,icl,ics):
     global pl1
     global sc1
     global p
@@ -1433,7 +1481,11 @@ def _parallel_init(ipl1x, isc1x, ipairs, ign1O, ign2O, iT2, iCORR, icorr_mode):
     global gn2O
     global T2
     global CORR
+    global cs
+    global cl
     global corr_mode
+    cs = ics
+    cl = icl
     pl1 = ipl1x
     sc1 = isc1x
     p = ipairs
@@ -1444,7 +1496,7 @@ def _parallel_init(ipl1x, isc1x, ipairs, ign1O, ign2O, iT2, iCORR, icorr_mode):
     corr_mode = icorr_mode
 
 @njit(parallel=True)
-def _refine_corr_kernel(p,indptr1,indices1,data1,indptr2,indices2,data2,n1,n2):
+def _refine_corr_kernel(filt, p,indptr1,indices1,data1,indptr2,indices2,data2,n1,n2):
     p1 = p[:,0]
     p2 = p[:,1]
     res = np.zeros(p1.size)
@@ -1468,13 +1520,26 @@ def _refine_corr_kernel(p,indptr1,indices1,data1,indptr2,indices2,data2,n1,n2):
         if izf.sum()>0:
             x=x[iz]
             y=y[iz]
-            res[j] = ((x-x.mean())*(y-y.mean()) / x.std() / y.std()).sum() / x.size
+            res[j] = ((x-x.mean())*(y-y.mean()) / x.std() / y.std())[filt[iz]].sum() / x.size
         else:
             res[j] = 0
             
     return res
             
-        
+def _xicorr(X,Y):
+    n = X.size
+    xi = np.argsort(X,kind='quicksort')
+    Y = Y[xi]
+    _,b,c = np.unique(Y,return_counts=True,return_inverse=True)
+    r = np.cumsum(c)[b]
+    _,b,c = np.unique(-Y,return_counts=True,return_inverse=True)
+    l = np.cumsum(c)[b]
+    denominator = (2*(l*(n-l)).sum())
+    if denominator > 0:
+        return 1 - n*np.abs(np.diff(r)).sum() / denominator
+    else:
+        return 0        
+
 def _refine_corr_parallel(
     sam1,
     sam2,
@@ -1488,6 +1553,7 @@ def _refine_corr_parallel(
     T1=0.0,
     T2=0.0,
     ncpus=os.cpu_count(),
+    ct_labels=None    
 ):
 
     import scipy as sp
@@ -1533,13 +1599,22 @@ def _refine_corr_parallel(
     p = pairs
     pl1 = pl1x
     sc1 = sc1x
-
-    if corr_mode == 'mutual_info':
+    
+    if corr_mode == 'mutual_info' or corr_mode == "xicorr":
         from multiprocessing import Pool, Manager
         CORR = Manager().dict()
         pc_chunksize = pl1.shape[1] // ncpus + 1
+        if ct_labels is None:
+            cl=cs=None
+        else:
+            cl,cs = ct_labels
+            if not (isinstance(cs,list) or isinstance(cs,tuple)):
+                cs = [cs]           
+            if np.in1d(cl,cs).sum()==0:
+                raise ValueError('Cell types not found in provided labels.')
+                
         pool = Pool(
-            ncpus, _parallel_init, [pl1, sc1, p, gn1O, gn2O, T2, CORR, corr_mode]
+            ncpus, _parallel_init, [pl1, sc1, p, gn1O, gn2O, T2, CORR, corr_mode, cl, cs]
         )
         try:
             pool.map(_parallel_wrapper, range(p.shape[0]), chunksize=pc_chunksize)
@@ -1548,7 +1623,17 @@ def _refine_corr_parallel(
             pool.join()
         CORR = CORR._getvalue()
     else:
-        vals = _refine_corr_kernel(p,pl1.indptr,pl1.indices,pl1.data,sc1.indptr,sc1.indices,sc1.data,pl1.shape[0],sc1.shape[0])
+        if ct_labels is not None:
+            cl,cs = ct_labels
+            if not (isinstance(cs,list) or isinstance(cs,tuple)):
+                cs = [cs]                      
+            filt = np.in1d(cl,cs)
+            if filt.sum()==0:
+                raise ValueError('Cell types not found in provided labels.')
+                
+        else:
+            filt = np.array([True]*pl1.shape[0])       
+        vals = _refine_corr_kernel(filt, p,pl1.indptr,pl1.indices,pl1.data,sc1.indptr,sc1.indices,sc1.data,pl1.shape[0],sc1.shape[0])
         CORR = dict(zip(to_vn(np.vstack((gn1O[p[:,0]],gn2O[p[:,1]])).T),vals))
     
     
@@ -1587,15 +1672,22 @@ try:
 except:
     pass;
 
-def hist2d(X,Y,bins=100):
+def hist2d(X,Y,bins=100,domain=None):
+    if domain is None:
+        xmin = X.min()
+        xmax = X.max()
+        ymin=Y.min()
+        ymax=Y.max()    
+        domain = [(xmin,xmax),(ymin,ymax)]
+    return histogram2d(X,Y,bins,domain)
+
+def calc_MI(X,Y,bins=100,cl=None,cs=None):    
     xmin = X.min()
     xmax = X.max()
     ymin=Y.min()
     ymax=Y.max()    
-    return histogram2d(X,Y,bins,[(xmin,xmax),(ymin,ymax)])
-
-def calc_MI(X,Y,bins=100):
-    c_XY = hist2d(X,Y,bins=bins)
+    domain = [(xmin,xmax),(ymin,ymax)]
+    c_XY = hist2d(X,Y,bins=bins,domain=domain)
     c_X = c_XY.sum(1)
     c_Y = c_XY.sum(0)
     
@@ -1637,16 +1729,10 @@ def _parallel_wrapper(j):
     ha = gn1O[j1] + ";" + gn2O[j2]
 
     try:
-        if corr_mode == 'pearson':
-            iz = np.logical_or(x>0,y>0)
-            izf = np.logical_and(x>0,y>0)
-            if izf.sum()>0:
-                CORR[ha] = np.corrcoef(x[iz],y[iz])[0,1]
-            else:
-                CORR[ha] = 0
-
+        if corr_mode == 'xicorr':
+            CORR[ha] = _xicorr(x,y)
         elif corr_mode == 'mutual_info':
-            CORR[ha] = calc_MI(x,y)
+            CORR[ha] = calc_MI(x,y,cl=cl,cs=cs)
         else:
             raise ValueError(f'`{corr_mode}` not recognized.')
     except:
